#!/usr/bin/python

from __future__ import division
import os
import random
import math
import shutil

try:
    from numpy import *
except ImportError, msg:
    print "[*] You need to install the numpy package before you can use morriscount!", `msg`
    sys.exit(1)

try:
    import h5py
except ImportError, msg:
    print "[*] You need to install the h5py package before you can use morriscount!", `msg`
    sys.exit(1)

try:
    import murmur
except ImportError, msg:
    print "[*] You need to install the 'murmur' package before you can use morriscount!", `msg`
    sys.exit(1)

serialize_attrs = [
'k','tokens_seen','vectors_seen','base','hashbits',
]

class MorrisCount(object):
    """
    Example code that implements a scalable approximate observation counting
    scheme that indexes Morris approximate counters in a Bloom filter to provide
    rough feature frequency tracking for an unbounded number of observations
    over an unboundedly high-dimensional feature space.
	
    Example application of this code:
    1) Count the number of packets you've seen from every network tuple on your network
    (src ip, src port, dest ip, dest port, proto).  Then record whenever you see a new or
    rare tuple to try to detect anomalies.

    2) Count the number of times you've seen each unique username say something on an
    IRC server, in order to flag the emergence of new handles.

    In other words, the intent here is to allow the user to track feature frequencies
    in high dimensional spaces typical of the security domain: for example,
    byte n-grams for network traffic or printable string observations within malware.

    As an example to show usage:
    To count the integers between 0 and 1e5-1, we would do the following:

    mc = MorrisCount()
    for i in range(100000):
        mc.add(i)
    mc.save("integers.h5")

    To use the class to retrieve counts, do:
    mycounter = MorrisCount.getcounter("integers.h5")

    Note:
    To use this code you need h5py, numpy, and murmur, all available via 'pip install'.

    Compatibility:
    Known to work on Ubuntu Linux and Mac OSX but likely will work anywhere Python
    is installed and h5py, murmur, and numpy are present.

    Please send questions to:
    josh.saxe@invincea.com
    """
    def __init__(self,hashbits=28,base=1.09,k=7):
        """
        Bloom filter will have hashbits**2 bytes; the bigger
        the filter the more accurate the counter.
        base controls the accuracy of the counts; a higher
        base means less accurate count tracking but will
        allow the counter to track higher numbers of observations.
        For example, a base of 2 will allow us to count to
        2**256.  1.09 will allow us to count to about 4 billion
        and is a good setting as it is quite accurate in lower
        ranges.  k is the traditional Bloom filter 'k' parameter
        and should be set using Bloom filter optimization math
        based on the expected size of the feature space you'll
        be counting.
        """
        self.k = k
        self.tokens_seen = 0
        self.vectors_seen = 0
        self.base = base
        self.counts = zeros(2**hashbits,dtype=uint8)
        self.hashbits = hashbits
    def log_prob(self, value):
        """Log probability of observing a feature given previously counted frequencies"""
        count = self.count(value)
        prob = min(1,(count+1)/ (self.vectors_seen+1))
        return math.log(prob,2)

    def prob(self, value):
        """Probability of observing a feature given previously counted frequencies"""
        count = self.count(value)
        prob = min(1,(count+1)/ (self.vectors_seen+1))
        if not self.vectors_seen:
            return 0
        else:
            prob = count / self.vectors_seen
            return prob

    def get_hashes(self,value):
        hvals = []
        for i in range(self.k):
            hval = murmur.string_hash(`value`,i)
            hval %= 2**self.hashbits
            hvals.append(hval)
        return hvals

    def add_vector(self,vector):
        """Add a set of observations to the counter"""
        for vec in vector:
            self.add(vec)
        self.vectors_seen += 1

    def add(self,value):
        try:
            value = unicode(value)
        except:
            pass
        hvals = self.get_hashes(value)
        exponent = min([self.counts[hval] for hval in hvals])
        if random.random() < (1.0/self.base)**exponent:
            for hval in hvals:
                self.counts[hval] += 1
        self.tokens_seen += 1

    def count(self,value):
        """How many times, approximately, have we seen this value?"""
        try:
            hvals = self.get_hashes(unicode(value))
        except:
            hvals = self.get_hashes(value)
        count = min([self.counts[hval] for hval in hvals])
        est = self.base**count-self.base
        est = est / (self.base - 1)
        return est + 1.0

    def count_items(self,items):
        """How many times, approximately, have we seen these values?"""
        results = []
        for i in items:
            results.append((i,self.count(i)))
        return dict(results)

    def log_prob_items(self,items):
        """Log probability of a set of items"""
        results = []
        for i in items:
            results.append((i,self.log_prob(i)))
        return dict(results)

    def save(self,name="morriscounts.h5"):
        """Save the counts to disk"""
        fpath = os.path.abspath(__file__)
        fpath = list(os.path.split(fpath)[:-1])
        fpath.append(name)
        fpath = "/".join(fpath)
        if os.path.exists(fpath):
            os.unlink(fpath)
        h5file = h5py.File(fpath, "w")
        h5file.create_dataset("bloom",data=self.counts)
        serialize_attrs = [
        'k','tokens_seen','vectors_seen','base','hashbits',
        ]
        for attr in serialize_attrs:
            h5file["bloom"].attrs[attr] = self.__dict__[attr]
        h5file.close()

def getcounter(name="morriscounts.h5"):
    """Get a counter instance"""
    fpath = os.path.abspath(__file__)
    fpath = list(os.path.split(fpath)[:-1])
    fpath.append(name)
    fpath = "/".join(fpath)
    mcounter = MorrisCount()
    if not os.path.exists(fpath):
        return mcounter
    h5file = h5py.File(fpath, "r")
    mcounter.counts = array(h5file["bloom"],dtype=uint8)
    for attr in serialize_attrs:
        mcounter.__dict__[attr] = h5file["bloom"].attrs[attr]
    h5file.close()
    return mcounter

def delcounter():
    """Delete a counter instance"""
    fpath = os.path.abspath(__file__)
    fpath = list(os.path.split(fpath)[:-1])
    fpath.append("morriscounts.h5")
    fpath = "/".join(fpath)

    spath = os.path.abspath(__file__)
    spath = list(os.path.split(spath)[:-1])
    spath.append("base_morriscounts.h5")
    spath = "/".join(spath)

    if os.path.exists(spath):
        shutil.copy(spath,fpath)
    elif not os.path.exists(spath) and os.path.exists(fpath):
        os.unlink(fpath)

if __name__ == '__main__':
    """Example usage of the MorrisCount class"""
    mc = MorrisCount()
    xs = []
    ys = []
    from matplotlib import pyplot as plt
    for i in range(100000):
        mc.add("hello")
        print i, mc.count("hello")

